/*-------------------------------------------------------------------------
	Copyright (C) 2008-20013 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"
#include "i2c-arbitration.h"

#define DEBUG
//#define DEBUG_VERBOSE

//#define DEBUG_SFPGA
#define AVOID_ERROR_LOG_SPAM

/* Fast timestamp variables used for debug purposes */
unsigned long start_req_jiffies=0, last_req_jiffies=0, last_grant_jiffies=0, last_release_jiffies=0, last_sst_jiffies[3]={ 0 };

/* Shared Bus Access functions */
int i2c_request_shared_bus(u8 segnum);
int i2c_grant_shared_bus(u8 segnum);
int i2c_release_shared_bus(u8 segnum);

int fpga_start_short_timer(void);
int fpga_set_local_short_timeout(u8 timeout);
int fpga_set_local_long_timeout(u8 timeout);

#ifdef DEBUG_SFPGA
void dump_support_fpga_regs(void);
#endif

dev_status_e
supfpga_mm_reg8_read(u16 reg, u8 *read_buf) {
	
	*read_buf = (u8)READ_SUPFPGA_MMAP_REG(reg);
	return STATUS_SUCCESS;
}

dev_status_e
supfpga_mm_reg16_read(u16 reg, u16 *read_buf) {
	
	*read_buf = (u16)READ_SUPFPGA_MMAP_REG(reg);
	return STATUS_SUCCESS;
}

dev_status_e
supfpga_mm_reg8_write(u16 reg, u8 write_buf) {

	WRITE_SUPFPGA_MMAP_REG(reg, (u16)write_buf);
	return STATUS_SUCCESS;
}

void
set_shared_bus_stats(u8 segnum, u8 val_mler, u8 val_mrer) {
	
	if (val_mler & MLER_LOC_REL_SHORT_TMO)
		g_sharedbus_stats[segnum].l_rel_short_timeout++;
		
	if (val_mler & MLER_LOC_REL_LONG_TMO)
		g_sharedbus_stats[segnum].l_rel_long_timeout++;
		
	if (val_mler & MLER_RM_GRT_TMO)
		g_sharedbus_stats[segnum].l_rm_grant_timeout++;
		
	if (val_mler & MLER_RM_REL_TMO)
		g_sharedbus_stats[segnum].l_rm_rel_timeout++;
		
	if (val_mler & MLER_LOC_SER_EXCH_ERR)
		g_sharedbus_stats[segnum].l_serial_exchange_error++;
		
	if (val_mrer & MRER_RM_REL_SHORT_TMO)
		g_sharedbus_stats[segnum].r_rel_short_timeout ++;
		
	if (val_mrer & MRER_RM_REL_LONG_TMO)
		g_sharedbus_stats[segnum].r_rel_long_timeout ++;
		
	if (val_mrer & MRER_LOC_GRT_TMO)
		g_sharedbus_stats[segnum].r_loc_grant_timeout ++;
		
	if (val_mrer & MRER_LOC_REL_TMO)
		g_sharedbus_stats[segnum].r_loc_rel_timeout ++;
		
	if (val_mrer & MLER_LOC_SER_EXCH_ERR)
		g_sharedbus_stats[segnum].r_serial_exchange_error ++;	
}

int 
i2c_request_shared_bus(u8 segnum) {

	u8 val_mlbr, val_mrbr, val_mler, val_mrer;
	dev_status_e retval=STATUS_SUCCESS;

#ifdef DEBUG_VERBOSE
	printk(KERN_INFO "i2c_request_shared_bus() ENTER: segnum=%d jiffies=%lu\n", (int)segnum, jiffies);
#endif

	start_req_jiffies = jiffies;

	/* set local long timeout value */
	fpga_set_local_long_timeout(4); 
	/* set short timer */
	fpga_set_local_short_timeout(255);
	/* start short timer */
	fpga_start_short_timer();

	/* read local bus register (MLBR) */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* read remote bus register (MRBR) for debugging */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);

	/* if local bus request already asserted (MLBR[4] set) */
	if (val_mlbr & MLBR_LOC_BUS_REQ) {
		/* log warning */
		printk(KERN_WARNING "  i2c_request_shared_bus(): Local bus request already asserted, segnum=%d, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
			(int)segnum, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);

		/* clear local bus request */
		supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (val_mlbr & ~MLBR_LOC_BUS_REQ));

		/* again read local bus request */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

		/* if local bus request is still asserted (MLBR[4] set) */
		if (val_mlbr & MLBR_LOC_BUS_REQ) {
			/* log error and return */
			printk(KERN_ERR "i2c_request_shared_bus() FAIL: Local bus request still asserted after deasserting, segnum=%d, MLBR=0x%02x, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
				(int)segnum, val_mlbr, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);
			return STATUS_INACCESSIBLE;
		} 
	}

	/* read local error register (MLER) */
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	/* read remote error register (MRER) */
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
	
	/* if any local or remote errors */
	if ((val_mler & CLEARABLE_ERROR_MASK) || (val_mrer & CLEARABLE_ERROR_MASK)) {
		/* log errors */
		printk(KERN_WARNING "  i2c_request_shared_bus(): Local and/or remote errors detected... attempting to clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);

		/* Set shared bus statistics when local and/or remote errors detected */
		set_shared_bus_stats(segnum, val_mler, val_mrer);

		/* if any local errors */
		if (val_mler & CLEARABLE_ERROR_MASK) {
			/* clear errors from local error register (note: clear is done be setting bit) */
			supfpga_mm_reg8_write(SUPFPGA_MLER_REG, val_mler);
			/* reread local error register (MLER) */
			supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
		}

		/* if any remote errors */
		if (val_mrer & CLEARABLE_ERROR_MASK) {
			/* clear errors from remote error register (note: clear is done be setting bit) */
			supfpga_mm_reg8_write(SUPFPGA_MRER_REG, val_mrer);
			/* reread remote error register (MRER) */
			supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
		}
			
		/* if we still have local or remote errors after clearing */
		if ((val_mler != NO_ERROR) || (val_mrer != NO_ERROR)) {
			printk(KERN_WARNING "i2c_request_shared_bus() FAIL: Local and/or remote errors detected after clearing, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
				(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);

			/* Set shared bus statistics when local and/or remote errors detected */
			set_shared_bus_stats(segnum, val_mler, val_mrer);

			return STATUS_INACCESSIBLE;
		}
	}

	/* if local bus already active (MLBR[5]set) */
	if (val_mlbr & MLBR_LOC_BUS_ACT) {
		printk(KERN_WARNING "i2c_request_shared_bus() FAIL: Local bus already active, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);
		return STATUS_INACCESSIBLE;
	} 

	/* set request bus number (segment number) */
	/* set the access type to read/write */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (segnum | MLBR_LOC_ACC_TYPE));
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* request bus */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (MLBR_LOC_BUS_REQ | val_mlbr));

	if (retval == STATUS_SUCCESS) {
#ifdef DEBUG_VERBOSE	
		/* read local bus register (MLBR) */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
		printk(KERN_INFO "i2c_request_shared_bus() SUCCESS: segnum=%d, MLBR=0x%02x, MRBR=0x%02x, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);
#endif
		last_req_jiffies = jiffies;
	}
	return retval;
}

int 
i2c_grant_shared_bus(u8 segnum) {

	u8 val_mlbr, val_mrbr, val_mler, val_mrer, long_timeout;
	static u8 last_mler=0, last_mrer=0;
	int timeout=0;
	int log_error=1; 
	dev_status_e retval=STATUS_SUCCESS;
	unsigned long start_jiffies, timeout_jiffies, loop_iterations=0;

	/* read local bus register (MLBR) */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	/* read remote bus register (MRBR) */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);

	/* read long timeout value */
	supfpga_mm_reg8_read(SUPFPGA_MLTR_REG, &long_timeout);

	/* for debugging purposes save start time */
	start_jiffies = jiffies;
	/* timeout in jiffies - (current time + long timer timeout) */	
	timeout_jiffies = (jiffies + msecs_to_jiffies(long_timeout * 1000));

#ifdef DEBUG_VERBOSE	
	printk(KERN_INFO "i2c_grant_shared_bus() ENTER: segnum=%d start_jiffies=%lu, timeout_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu\n", 
		(int)segnum, start_jiffies, timeout_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies);
#endif

	/* while bus is not active (MLBR[5] clear) */
	while ((0 == (val_mlbr & MLBR_LOC_BUS_ACT))) {

		loop_iterations++;
		
		/* if timeout */
		if (time_after(jiffies, timeout_jiffies)) {
			timeout = 1;
			break;
		}

		/* pet short timer */
		fpga_start_short_timer();
		
		/* Read local errors(MLER) and remote errors(MRER) */
		supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
		supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
		
		/* if any error detected (local, remote, clearable, unclearable) */
		if ((val_mler != NO_ERROR) || (val_mrer != NO_ERROR)) {

#ifdef AVOID_ERROR_LOG_SPAM
			/* if un-clearable errors only (invalid serial exchange bus AND local buffers not ready)*/
			if (((val_mler & CLEARABLE_ERROR_MASK) == NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) == NO_ERROR)) {

				 /* Some errors are not clearable and can potentially cause log spam.  
				 Avoid log spam by only logging once everytime error asserts */
				if ( /*((val_mler & 0x80) && (last_mler & 0x80)) || RESERVED */ 
					 ((val_mrer & 0x80) && (last_mrer & 0x80)) ||
					 ((val_mler & 0x40) && (last_mler & 0x40)) ||
					 ((val_mrer & 0x40) && (last_mrer & 0x40)) ||
					 ((val_mler & 0x20) && (last_mler & 0x20)) ||
					 ((val_mrer & 0x20) && (last_mrer & 0x20)) ) {
						log_error = 0; /* this error has already been logged, so skip logging */
				}
			}
#endif
			/* log the error (to avoid log spam from un-clearable errors 
			we only log if error has changed since last loop) */
			if (log_error) {
				printk(KERN_WARNING "  i2c_grant_shared_bus(): Local and/or remote errors... attempting to clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, timeout_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, timeout_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);
			}

#ifdef DEBUG_SFPGA
			dump_support_fpga_regs();
			retval = STATUS_INACCESSIBLE;
			break; /* note: request will be released before we return from function */
#endif
			
			/* release bus before attempting to clear errors */
			i2c_release_shared_bus(segnum);

			/* if mler errors attempt to clear */
			if ((val_mler & CLEARABLE_ERROR_MASK) != NO_ERROR) {
				supfpga_mm_reg8_write(SUPFPGA_MLER_REG, val_mler);
				supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
			}
			/* if mrer errors attempt to clear */
			if ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR) {
				supfpga_mm_reg8_write(SUPFPGA_MRER_REG, val_mrer);
				supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
			}

			/* re-request bus request */
			if (i2c_request_shared_bus(segnum) != 0) {
				printk(KERN_WARNING "  i2c_grant_shared_bus(): Re-request of bus after error detection failed, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, timeout_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, timeout_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);
				retval = STATUS_INACCESSIBLE;
				break;
			}

			/* if still have errors... log error */
			if (((val_mler & CLEARABLE_ERROR_MASK) != NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR)) {
			
				printk(KERN_WARNING "  i2c_grant_shared_bus(): Local and/or remote errors after clear, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, timeout_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
					(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, timeout_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);

				/* Set shared bus statistics when local and/or remote errors detected */
				set_shared_bus_stats(segnum, val_mler, val_mrer);

				retval = STATUS_INACCESSIBLE;
				break;
			}
		}

		last_mler = val_mler;
		last_mrer = val_mrer;

		/* allow kernel to do other stuff while waiting for bus grant */
		schedule_timeout_uninterruptible(1); /* approx 4ms sleep */
		
		/* re-read MLBR and MRBR registers (update val_mlbr and val_mrbr) */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
		supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	}
 
	/* if wait timer >= long timeout value */
	if (timeout) {
#ifdef DEBUG
		printk(KERN_WARNING "  i2c_grant_shared_bus(): Timeout waiting for bus grant, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, timeout_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, timeout_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);
#endif	
		retval = STATUS_INACCESSIBLE;
	}

	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);

	/* release bus request on failures */
	if (retval != STATUS_SUCCESS) {
#ifdef DEBUG_VERBOSE	
		printk(KERN_WARNING "i2c_grant_shared_bus() FAIL, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);
#endif			
		i2c_release_shared_bus(segnum);
	}

	if (retval == STATUS_SUCCESS) {
		fpga_start_short_timer();
#ifdef DEBUG_VERBOSE	
		printk(KERN_INFO "i2c_grant_shared_bus() SUCCESS, segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, start_jiffies=%lu, start_req_jiffies=%lu, last_req_jiffies=%lu, last_grant_jiffies=%lu, last_release_jiffies=%lu, last_sst_jiffies=%lu %lu %lu, jiffies=%lu, loop_iterations=%lu\n", 
			(int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, start_jiffies, start_req_jiffies, last_req_jiffies, last_grant_jiffies, last_release_jiffies, last_sst_jiffies[0], last_sst_jiffies[1], last_sst_jiffies[2], jiffies, loop_iterations);
#endif	
		last_grant_jiffies = jiffies;
	} 
	
	return retval;
}

int
i2c_release_shared_bus(u8 segnum) {

	u8 val_mlbr;
#ifdef DEBUG_VERBOSE	
	u8 val_mrbr, val_mler, val_mrer;
#endif

	/* read local bus register */
	supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

	last_release_jiffies = jiffies;

	/* release bus request (clear MLBR[4]) */
	supfpga_mm_reg8_write(SUPFPGA_MLBR_REG, (val_mlbr & ~MLBR_LOC_BUS_REQ));	

#ifdef DEBUG_VERBOSE
	/* log any errors (we will clear them on next access) */
	supfpga_mm_reg8_read(SUPFPGA_MRBR_REG, &val_mrbr);
	supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
	supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
	if (((val_mler & CLEARABLE_ERROR_MASK)!= NO_ERROR) || ((val_mrer & CLEARABLE_ERROR_MASK) != NO_ERROR)) {
		printk(KERN_INFO "i2c_release_shared_bus(): segnum=%d, MLBR=0x%02x, MRBR=0x%02x, MLER=0x%02x, MRER=0x%02x, jiffies=%lu\n", (int)segnum, val_mlbr, val_mrbr, val_mler, val_mrer, jiffies);
	}
#endif

	return STATUS_SUCCESS;
}


int 
fpga_start_short_timer() {

	static int cnt=0;
	last_sst_jiffies[cnt] = jiffies;
	(cnt == 2) ? cnt = 0 : cnt++;

	supfpga_mm_reg8_write(SUPFPGA_MTRR_REG, MTRR_START_STIMER);

	return STATUS_SUCCESS;
}

int 
fpga_set_local_short_timeout(u8 timeout) {
	
#ifdef DEBUG_VERBOSE	
	printk(KERN_INFO "fpga set local short timeout to %d millisecond(s)\n", (int)timeout);
#endif
	supfpga_mm_reg8_write(SUPFPGA_MSTR_REG, timeout);

	return STATUS_SUCCESS;
}

int
fpga_set_local_long_timeout(u8 timeout) {
	
	/* max allowed long timeout value is 15 */
	if (timeout > 15) {
#ifdef DEBUG_VERBOSE	
		printk(KERN_WARNING "fpga set local long timeout to %d second(s) exceeds max of 15 seconds\n   forcing timeout to 15 seconds\n", (int)timeout);
#endif
		timeout = 15;
	}
	
#ifdef DEBUG_VERBOSE	
	printk(KERN_INFO "fpga set local long timeout to %d second(s)\n", (int)timeout);
#endif
	supfpga_mm_reg8_write(SUPFPGA_MLTR_REG, timeout);

	return STATUS_SUCCESS;
}

#ifdef DEBUG_SFPGA
void
dump_support_fpga_regs() {

	u16 val, usw;

	/* log support fpga registers */
	printk(KERN_WARNING "  SUPPORT FPGA REG DUMP: \n");

	supfpga_mm_reg16_read(SUPFPGA_MCSR_REG, &val);
	printk(KERN_WARNING "    MCSR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLBR_REG, &val);
	printk(KERN_WARNING "    MLBR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MTRR_REG, &val);
	printk(KERN_WARNING "    MTRR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MRBR_REG, &val);
	printk(KERN_WARNING "    MRBR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLER_REG, &val);
	printk(KERN_WARNING "    MLER=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MRER_REG, &val);
	printk(KERN_WARNING "    MRER=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MSTR_REG, &val);
	printk(KERN_WARNING "    MSTR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_MLTR_REG, &val);
	printk(KERN_WARNING "    MLTR=0x%04x\n", val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG0_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG0_REG-2, &usw);
	printk(KERN_WARNING "    C000=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG1_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG1_REG-2, &usw);
	printk(KERN_WARNING "    C004=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG2_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG2_REG-2, &usw);
	printk(KERN_WARNING "    C008=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG3_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG3_REG-2, &usw);
	printk(KERN_WARNING "    C00C=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG4_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG4_REG-2, &usw);
	printk(KERN_WARNING "    C010=0x%04x%04x\n", usw, val);

	supfpga_mm_reg16_read(SUPFPGA_DEBUG5_REG, &val);
	supfpga_mm_reg16_read(SUPFPGA_DEBUG5_REG-2, &usw);
	printk(KERN_WARNING "    C014=0x%04x%04x\n", usw, val);
	
	return;
}			
#endif
